"""
Created on Sun Jul  7 17:42:28 2019

@authors: Pranay Barkataki 

Nth neighbor inhomogeneous spin interaction Hamiltonian with periodic boundary 
condition
"""
import numpy as np
from QuantumInformation import QuantumMechanics as QM
qobj=QM()

class Hamiltonian:
    
    def field_xham(self,N,mode='homogeneous',B=1):
        """
        Constructs Hamiltonian of external magnetic field in X direction
        Input
            N: number of spins
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            B: it list of value if mode='inhomogeneous', and constant if
               mode='homogeneous' 
        Output
            xham: Hamiltonian
        """
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        if mode == 'homogeneous':
            xham=np.zeros([2**N,2**N],dtype='float64')
            for i in range(0,2**N,1):
                for j in range(0,2**N,1):
                    bvec=qobj.decimal_binary(j,N)
                    sum1=0.0
                    for k in range(0,N):
                        bb=np.copy(bvec)
                        bb[k]=1-bb[k]
                        row=qobj.binary_decimal(bb)
                        if row == i:
                            sum1=sum1+B
                    xham[i,j]=sum1                             
        else:
            assert len(B)==N,\
            "The entered values of magnetic strengths are not equal to number of spins"
            xham=np.zeros([2**N,2**N],dtype='float64')
            for i in range(0,2**N,1):
                for j in range(0,2**N,1):
                    bvec=qobj.decimal_binary(j,N)
                    sum1=0.0
                    for k in range(0,N):
                        bb=np.copy(bvec)
                        bb[k]=1-bb[k]
                        row=qobj.binary_decimal(bb)
                        if row == i:
                            sum1=sum1+B[k]
                    xham[i,j]=sum1
        return xham
    
    def field_yham(self,N,mode='homogeneous',B=1):
        """
        Constructs Hamiltonian of external magentic field in Y direction
        Input
            N: number of spins
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            B: it list of value if mode='inhomogeneous', and constant if
               mode='homogeneous' 
        Output
            yham: Hamiltonian
        """
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        if mode == 'homogeneous':
            yham=np.zeros([2**N,2**N],dtype=np.complex_)
            for i in range(0,2**N,1):
                for j in range(0,2**N,1):
                    bvec=qobj.decimal_binary(j,N)
                    sum1=0.0
                    for k in range(0,N):
                        bb=np.copy(bvec)
                        bb[k]=1-bb[k]
                        row=qobj.binary_decimal(bb)
                        if row == i:
                            sum1=sum1+(B*complex(0,1)*((-1)**bvec[k]))
                    yham[i,j]=sum1                             
        else:
            assert len(B)==N,\
            "The entered values of magnetic strengths are not equal to number of spins"
            yham=np.zeros([2**N,2**N],dtype=np.complex_)
            for i in range(0,2**N,1):
                for j in range(0,2**N,1):
                    bvec=qobj.decimal_binary(j,N)
                    sum1=0.0
                    for k in range(0,N):
                        bb=np.copy(bvec)
                        bb[k]=1-bb[k]
                        row=qobj.binary_decimal(bb)
                        if row == i:
                            sum1=sum1+(B[k]*complex(0,1)*((-1)**bvec[k]))
                    yham[i,j]=sum1
        return yham
    
    def field_zham(self,N,mode='homogeneous',B=1):
        """
        Constructs Hamiltonian of external magentic field in Z direction
        Input
            N: number of spins
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            B: it list of value if mode='inhomogeneous', and constant if
               mode='homogeneous' 
        Output
            zham: Hamiltonian
        """
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        zham=np.zeros([2**N,2**N],dtype='float64')
        if mode == 'homogeneous':
            for i in range(0,2**N):
                sum1=0.0
                bvec=qobj.decimal_binary(i,N)
                for k in range(0,N):
                    sum1=sum1+(((-1)**bvec[k])*B)
                zham[i,i]=sum1
        else:
            assert len(B)==N,\
            "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                sum1=0.0
                bvec=qobj.decimal_binary(i,N)
                for k in range(0,N):
                    sum1=sum1+(((-1)**bvec[k])*B[k])
                zham[i,i]=sum1
        return zham     

    def ham_xx(self,N,nn=1,mode='homogeneous',jx=1,condition='periodic'):
        """
        Constructs Hamiltonian of spin-spin interaction in X direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            jx: it list of values if mode='inhomogeneous', and constant if
                 mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian
        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        if mode == 'homogeneous':
            ham=np.zeros([2**N,2**N],dtype='float64')
            col2=np.zeros([N,1])
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jx)
        else:
            if condition=='periodic':
                assert len(jx)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(jx)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            ham=np.zeros([2**N,2**N],dtype='float64')
            col2=np.zeros([N,1])
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jx[k])
        return ham               
    
    def ham_yy(self,N,nn=1,mode='homogeneous',jy=1,condition='periodic'):
        """
        Constructs Hamiltonian of spin-spin interaction in Y direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            jy: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian
        
        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype='float64')
        col2=np.zeros([N,1])
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        if col2[k]==col2[k1]:
                            ind=-1
                        else:
                            ind=1
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jy*ind)
        else:
            if condition=='periodic':
                assert len(jy)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(jy)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        if col2[k]==col2[k1]:
                            ind=-1
                        else:
                            ind=1
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jy[k]*ind)
        return ham
        
    def ham_zz(self,N,nn=1,mode='homogeneous',jz=1,condition='periodic'):
        """
        Constructs Hamiltonian of spin-spin interaction in Z direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            jz: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian
        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype='float64')
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        egv1=1-2*col[k]
                        egv1=int(egv1)
                        egv2=1-2*col[k1]
                        egv2=int(egv2)
                        if i==j:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jz*egv1*egv2)
        else:
            if condition=='periodic':
                assert len(jz)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(jz)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                for j in range(0,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        egv1=1-2*col[k]
                        egv1=int(egv1)
                        egv2=1-2*col[k1]
                        egv2=int(egv2)
                        if i==j:
                            inn=1
                        else:
                            inn=0
                        ham[i,j]=ham[i,j]+(inn*jz[k]*egv1*egv2)
        return ham
    
    def heisenberg_hamiltonian(self,N,nn=1,mode='homogeneous',j=1.0,\
                               condition='periodic'):
        """
        Construct Heisenberg interaction type Hamiltonian
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            j: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian

        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype='float64')
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for jj in range(0,2**N):
                    col=qobj.decimal_binary(jj,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        temp=col2[k]
                        col2[k]=col2[k1]
                        col2[k1]=temp
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,jj]=ham[i,jj]+(inn*j*2.0)
                        if i==jj:
                            ham[i,jj]=ham[i,jj]-j
        else:
            if condition=='periodic':
                assert len(j)==N,\
                "Entered list of values j are not equal to number of interaction in PBC"
            else:
                assert len(j)==N-nn,\
                "Entered list of values j are not equal to number of interaction in OBC"
            for i in range(0,2**N):
                for jj in range(0,2**N):
                    col=qobj.decimal_binary(jj,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        temp=col2[k]
                        col2[k]=col2[k1]
                        col2[k1]=temp
                        dec=qobj.binary_decimal(col2)
                        if dec==i:
                            inn=1
                        else:
                            inn=0
                        ham[i,jj]=ham[i,jj]+(inn*j[k]*2.0)
                        if i==jj:
                            ham[i,jj]=ham[i,jj]-j[k]
        return ham
    
    def dm_xham(self,N,nn=1,mode='homogeneous',dx=1.0, condition='periodic'):
        """
        Construct DM Hamiltonian in X direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            dx: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian

        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype=np.complex_)
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        dec=qobj.binary_decimal(col2)
                        inn1=complex(0.0,0.0)
                        if dec==i:                            
                            phase=complex(0,1)*(-1)**col[k]
                            szpart=(-1)**col[k1]
                            inn1=phase*szpart
                        col2=col.copy()
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn2=complex(0.0,0.0)
                        if dec==i:                            
                            phase=complex(0,1)*(-1)**col[k1]
                            szpart=(-1)**col[k]
                            inn2=phase*szpart
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dx)
                    ham[j,i]=np.conjugate(ham[i,j])
        else:
            if condition=='periodic':
                assert len(dx)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(dx)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        dec=qobj.binary_decimal(col2)
                        inn1=0.0
                        if dec==i:                            
                            phase=complex(0,1)*(-1)**col[k]
                            szpart=(-1)**col[k1]
                            inn1=phase*szpart
                        col2=col.copy()
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn2=0.0
                        if dec==i:                            
                            phase=complex(0,1)*(-1)**col[k1]
                            szpart=(-1)**col[k]
                            inn2=phase*szpart
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dx[k])
                    ham[j,i]=np.conjugate(ham[i,j])
        return ham
    
    def dm_yham(self,N,nn=1,mode='homogeneous',dy=1.0, condition='periodic'):
        """
        Construct DM Hamiltonian in Y direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            dy: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian

        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype='float64')
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn1=0.0
                        if dec==i:                            
                            inn1=(-1)**col[k]
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        dec=qobj.binary_decimal(col2)
                        inn2=0.0
                        if dec==i:                          
                            inn2=(-1)**col[k1]
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dy)
                    ham[j,i]=ham[i,j]
        else:
            if condition=='periodic':
                assert len(dy)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(dy)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn1=0.0
                        if dec==i:                            
                            inn1=(-1)**col[k]
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        dec=qobj.binary_decimal(col2)
                        inn2=0.0
                        if dec==i:                          
                            inn2=(-1)**col[k1]
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dy[k])
                    ham[j,i]=ham[i,j]
        return ham
        
    def dm_zham(self,N,nn=1,mode='homogeneous',dz=1.0, condition='periodic'):
        """
        Construct DM Hamiltonian in Z direction
        Input
            N: number of spins
            nn: specifying rth interaction
            mode: 'homogeneous' or 'inhomogeneous' magnetic field
            dz: it list of values if mode='inhomogeneous', and constant if
                mode='homogeneous'
            condition:  defining boundary conditions of the Hamiltonian
        Output
            ham: Hamiltonian

        """
        assert condition == 'periodic' or condition == 'open'
        if condition == 'periodic':
            kn=N
        else:
            kn=N-nn
        assert mode == 'homogeneous' or mode == 'inhomogeneous',\
        "Entered mode is invalid"
        assert N >= 1, "number of spins entered is not correct"
        assert nn<=N-1 and nn>=1,"Not valid interaction"
        ham=np.zeros([2**N,2**N],dtype=np.complex_)
        if mode == 'homogeneous':
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn1=complex(0.0,0.0)
                        inn2=complex(0.0,0.0)
                        if dec==i:                            
                            inn1=complex(0,1)*(-1)**col[k1]
                            inn2=complex(0,1)*(-1)**col[k]
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dz)
                    ham[j,i]=np.conjugate(ham[i,j])
        else:
            if condition=='periodic':
                assert len(dz)==N,\
                "The entered values of magnetic strengths are not equal to number of spins"
            else:
                assert len(dz)==N-nn,\
                "The entered values of magnetic strengths are not equal to number of spins"
            for i in range(0,2**N):
                for j in range(i,2**N):
                    col=qobj.decimal_binary(j,N)
                    for k in range(0,kn):
                        k1=k+nn
                        if k1>=N:
                            k1=k1-N
                        col2=col.copy()
                        col2[k]=1-col2[k]
                        col2[k1]=1-col2[k1]
                        dec=qobj.binary_decimal(col2)
                        inn1=complex(0.0,0.0)
                        inn2=complex(0.0,0.0)
                        if dec==i:                            
                            inn1=complex(0,1)*(-1)**col[k1]
                            inn2=complex(0,1)*(-1)**col[k]
                        ham[i,j]=ham[i,j]+((inn1-inn2)*dz[k])
                    ham[j,i]=np.conjugate(ham[i,j])
        return ham
        
        
        
        
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar  2 11:35:07 2021

@author: pranay barkataki
"""

import numpy as np
import math
import cmath
from QuantumInformation import RecurNum
from QuantumInformation import QuantumMechanics as QM
from QuantumInformation import LinearAlgebra as LA
import re

class GatesTools:
    
    def __init__(self):
        """It is a class dealing with Gates, entropy, and
        it primarily intrinsi functions of uses numpy, math, cmath.
        """
        
    def sx(self,N=1):
        """
        Construct N-qubit Pauli spin matrix sigma_x
        Inputs:
            N: number of spins
        Output:
            sigmax: It stores the Pauli spin matrix sx
        """
        sigmax=np.zeros([2**N,2**N])
        j=(2**N)-1
        for i in range(0,2**N):
            sigmax[i,j]=1
            j=j-1
        return sigmax
    
    def sy(self,N=1):
        """
        Construct N-qubit Pauli spin matrix sigma_y
        Inputs:
            N: Number of spins
        Outputs:
            sigmay: It stores the Pauli spin matrix sy
        """
        sigmay2=np.array([[0,complex(0,-1)],[complex(0,1),0]])
        if N >1:
            for i in range(2,N+1):
                if i==2:
                    sigmay=np.kron(sigmay2, sigmay2)
                elif i > 2:
                    sigmay=np.kron(sigmay, sigmay2)
        else:
            sigmay=sigmay2
        
        return sigmay
    
    def sz(self,N=1):
        """
        Construct N-qubit Pauli spin matrix sigma_z
        Inputs:
            N: Number of spins
        Outputs:
            sigmaz: It stores the Pauli spin matrix sz
        """
        sigmaz2=np.array([[1,0],[0,-1]])
        if N >1:
            for i in range(2,N+1):
                if i==2:
                    sigmaz=np.kron(sigmaz2, sigmaz2)
                elif i > 2:
                    sigmaz=np.kron(sigmaz, sigmaz2)
        else:
            sigmaz=sigmaz2
        
        return sigmaz
    
    def hadamard_mat(self,N=1):
        """
        Construct N-qubit Hadamard matrix
        Inputs:
            N: Number of spins
        Outputs:
            hadamard: It stores the Hadamard matrix
        """
        hadamard2=np.array([[1/np.sqrt(2),1/np.sqrt(2)],\
                           [1/np.sqrt(2),-1/np.sqrt(2)]])
        if N >1:
            for i in range(2,N+1):
                if i==2:
                    hadamard=np.kron(hadamard2, hadamard2)
                elif i > 2:
                    hadamard=np.kron(hadamard, hadamard2)
        else:
            hadamard=hadamard2
        
        return hadamard
    
    def phase_gate(self,N=1):
        """
        Construct N-qubit phase gate matrix
        Inputs:
            N: Number of spins
        Outputs:
            phaseg: It stores the phase gate matrix
        """
        phaseg2=np.array([[1,0],\
                           [0,complex(0,1)]])
        if N >1:
            for i in range(2,N+1):
                if i==2:
                    phaseg=np.kron(phaseg2, phaseg2)
                elif i > 2:
                    phaseg=np.kron(phaseg, phaseg2)
        else:
            phaseg=phaseg2
        
        return phaseg
    
    def rotation_gate(self,k,N=1):
        """
        Input:
            k: is a positive number
            N: number of spins
        Returns:
            rotg: Rotation gate matrix
        """
        assert k > 0, "k is not positive number"  
        
        z=complex(0,(2*math.pi)/(2**k))
        rotg2=np.array([[1,0],[0,cmath.exp(z)]])
        if N >1:
            for i in range(2,N+1):
                if i==2:
                    rotg=np.kron(rotg2, rotg2)
                elif i > 2:
                    rotg=np.kron(rotg, rotg2)
        else:
            rotg=rotg2
        
        return rotg     
    
    def cx_gate(self):
        """
        It returns controlled NOT gate
        """
        return np.array([[1,0,0,0],[0,1,0,0],[0,0,0,1],[0,0,1,0]])
    
    def cz_gate(self):
        """
        It returns controlled Z gate
        """
        
        return np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,-1]])
    
    def swap_gate(self):
        """
        It returns a swap gate
        """
        
        return np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]])
    
    def toffoli_gate(self):
        """
        It returns a Toffoli gate
        """
        
        return np.array([[1,0,0,0,0,0,0,0],[0,1,0,0,0,0,0,0],\
                         [0,0,1,0,0,0,0,0],[0,0,0,1,0,0,0,0],\
                         [0,0,0,0,1,0,0,0],[0,0,0,0,0,1,0,0],\
                         [0,0,0,0,0,0,0,1],[0,0,0,0,0,0,1,0]])
    
    def fredkin_gate(self):
        """
        It returns a Fredkin gate
        """
        
        return np.array([[1,0,0,0,0,0,0,0],[0,1,0,0,0,0,0,0],\
                         [0,0,1,0,0,0,0,0],[0,0,0,1,0,0,0,0],\
                         [0,0,0,0,1,0,0,0],[0,0,0,0,0,0,1,0],\
                         [0,0,0,0,0,1,0,0],[0,0,0,0,0,0,0,1]])
    

    # RN and LN state having plus sign
    def bell1(self,tot_spins=2,shift=0):
        """
        Construct N tensor products of the |bell1> or T|bell1> Bell state
        Input:
            tot_spins: The total number of spins
            shift: for value 0 we get |bell1> and for value 1 T|bell1>.
        Output:
            state: the result will be |bell1> or T|bell1> state.
        """
        assert tot_spins%2==0, "the total number of spins is not an even number"
        assert shift==0 or shift==1, "Invalid entry of the shift value"
        terms=int(tot_spins/2)
        row=np.zeros([terms,1])
        mylist=[]
        icount=-1
        RecurNum.RecurChainRL1(row,tot_spins,icount,mylist,shift)
        mylist=np.array(mylist)
        state=np.zeros([2**tot_spins])
        factor=1/math.sqrt(2)
        len_mylist=len(mylist)
        for x in range(0,len_mylist):
            state[mylist.item(x)]=factor**terms
        return(state)
    

    # RN and LN state constructed from the singlet state
    def bell2(self,tot_spins=2,shift=0):
        """
        Construct N tensor products of the |bell2> or T|bell2> Bell state
        Input:
            tot_spins: The total number of spins
            shift: for value 0 we get |bell2> and for value 1 T|bell2>.
        Output:
            state: the result will be |bell2> or T|bell2> state.
        """
        assert tot_spins%2==0, "the total number of spins is not an even number"
        assert shift==0 or shift==1, "Invalid entry of the shift value"
        terms=int(tot_spins/2)
        row=np.zeros([terms,1])
        mylist=[]
        icount=-1
        RecurNum.RecurChainRL2(row,tot_spins,icount,mylist,shift)
        mylist=np.array(mylist)
        state=np.zeros([2**tot_spins])
        factor=1/math.sqrt(2)
        len_mylist=len(mylist)
        for x in range(0,len_mylist):
            if mylist.item(x)<0:
                state[-mylist.item(x)]=-factor**terms
            elif mylist.item(x)>=0:
                state[mylist.item(x)]=factor**terms
        return(state)
        
    # 00 and 11 bell state having plus sign
    def bell3(self,tot_spins=2,shift=0):
        """
        Construct N tensor products of the |bell3> or T|bell3> Bell state
        Input:
            tot_spins: The total number of spins
            shift: for value 0 we get |bell3> and for value 1 T|bell3>.
        Output:
            state: the result will be |bell3> or T|bell3> state.
        """
        assert tot_spins%2==0, "the total number of spins is not an even number"
        assert shift==0 or shift==1, "Invalid entry of the shift value"
        terms=int(tot_spins/2)
        row=np.zeros([terms])
        mylist=[]
        icount=-1
        RecurNum.RecurChainRL3(row,tot_spins,icount,mylist,shift)
        mylist=np.array(mylist)
        state=np.zeros([2**tot_spins])
        factor=1/math.sqrt(2)
        len_mylist=len(mylist)
        for x in range(0,len_mylist):
            state[mylist.item(x)]=factor**terms
        return(state)
    
    # 00 and 11 bell state having negative sign
    def bell4(self,tot_spins=2,shift=0):
        """
        Construct N tensor products of the |bell4> or T|bell4> Bell state
        Input:
            tot_spins: The total number of spins
            shift: for value 0 we get |bell4> and for value 1 T|bell4>.
        Output:
            state: the result will be |bell4> or T|bell4> state.
        """
        assert tot_spins%2==0, "the total number of spins is not an even number"
        assert shift==0 or shift==1, "Invalid entry of the shift value"
        terms=int(tot_spins/2)
        row=np.zeros([terms,1])
        mylist=[]
        icount=-1
        RecurNum.RecurChainRL4(row,tot_spins,icount,mylist,shift)
        mylist=np.array(mylist)
        state=np.zeros([2**tot_spins])
        factor=1/math.sqrt(2)
        len_mylist=len(mylist)
        for x in range(0,len_mylist):
            if mylist.item(x)<0:
                state[-mylist.item(x)]=-factor**terms
            elif mylist.item(x)>=0:
                state[mylist.item(x)]=factor**terms
        return(state)
    
    def nGHZ(self,tot_spins=3):
        """
        Construct N-qubit GHZ state
        Input:
            tot_spins: it is the total number of spins, it should be equal to 
                       or greater than 3.
        Output:
            state: N-qubit GHZ state.
        """
        assert tot_spins >= 3, "Total number of spins are less than 3"
        state=np.zeros([2**tot_spins])
        state[0]=1/np.sqrt(2)
        state[(2**tot_spins)-1]=1/np.sqrt(2)
        return state
    
    def nW(self,tot_spins=3):
        """
        Construct N-qubit W state
        Input:
            tot_spins: it is the total number of spins, it should be equal to 
                       or greater than 3.
        Output:
            state: N-qubit W state.
        """        
        assert tot_spins >= 3, "Total number of spins are less than 3"
        state=np.zeros([2**tot_spins])
        for i in range(0,tot_spins):
            state[2**i]=1/np.sqrt(tot_spins)
        return state
        
    def nWerner(self,p,tot_spins=2):
        """
        Construct N-qubit Werner state
        Input:
            tot_spins: it is the total number of spins, it should be equal to
                        or greater than 2.
            p: it is the mixing probability
        Output:
            rho: N-qubit Werner state.
        """
        assert tot_spins >= 2, "Total number of spins are less than 2"
        qobj=QM()
        if tot_spins == 2:
            state=self.bell3()
        else:
            state=self.nGHZ(tot_spins=tot_spins)
        den=qobj.outer_product_rvec(state,state)
        identity=np.identity(2**tot_spins, dtype = 'float64')
        identity=identity*(1/(2**tot_spins))
        rho=(p*den)+((1-p)*identity)
        return rho
    
    def shannon_entropy(self,pvec):
        """
        Calculates the Shannon entropy
        Input:
            pvec: column vector which contains probabilities
        Output:
            se: it returns the Shannon entropy value
        """
        size=pvec.shape[0]
        se=0.0
        for i in range(0,size):
            assert pvec[i]<=1 and pvec[i] >=0, "probability values are incorrect"
            se=se-(pvec[i]*math.log2(pvec[i]))
        return se
    
    def linear_entropy(self, rho):
        """
        Calculates the Linear entropy
        Input:
            rho: it is the density matrix
        Output:
            le: linear entropy value
        """
        tr=np.trace(rho)
        assert np.allclose(abs(tr),1), "density matrix is not correct"
        tr2=np.trace(np.matmul(rho,rho))
        le=1.0-abs(tr2)
        return le
    
    def relative_entropy(self,rho,sigma):
        """
        Calculates relative entropy
        Input:
            rho: input density matrix
            sigma: input density matrix
        Output:
            rtent: the value of relative entropy
        """
        laobj=LA()
        typerho=str(rho.dtype)
        typesig=str(sigma.dtype)
        if re.findall('^float|^int',typerho):
            logrho=laobj.function_smatrix(rho, mode="log",\
                                          log_base=math.exp(1))
        elif re.findall("^complex",typerho):
            logrho=laobj.function_hmatrix(rho, mode="log",\
                                          log_base=math.exp(1))
        if re.findall('^float|^int',typesig):
            logsig=laobj.function_smatrix(sigma, mode="log",\
                                          log_base=math.exp(1))
        elif re.findall("^complex",typesig):
            logsig=laobj.function_hmatrix(sigma, mode="log",\
                                          log_base=math.exp(1))    
        rtent=np.trace(np.matmul(rho,logrho))-np.trace(np.matmul(rho,logsig))
        rtent=abs(rtent)
        return rtent
    
    def trace_distance(self,rho,sigma):
        """
        Calculates trace distance between two density matrices
        Input:
            rho: input density matrix
            sigma: input density matrix
        Output:
            trd: it stores trace distance
        """
        res=rho-sigma
        laobj=LA()
        typeres=str(res.dtype)
        if re.findall('^float|^int',typeres):
            trd=laobj.trace_norm_rmatrix(res)
            trd=trd/2
        elif re.findall("^complex",typeres):
            trd=laobj.trace_norm_cmatrix(res)
            trd=trd/2
        return trd
    
    def fidelity_den2(self,rho,sigma):
        """
        Calculates fidelity between two density matrices
        Input:
            rho: input density matrix
            sigma: input density matrix
        Output:
            fidelity: it stores the value of fidelity
        """
        laobj=LA()
        typerho=str(rho.dtype)
        typesig=str(sigma.dtype)
        flag=0
        if re.findall('^float|^int',typerho):
            rhosq=laobj.power_smatrix(rho,0.5)
        elif re.findall("^complex",typerho):
            rhosq=laobj.power_hmatrix(rho,0.5)
            flag=1
        if re.findall('^float|^int',typesig):
            sigsq=laobj.power_smatrix(sigma,0.5)
        elif re.findall("^complex",typesig):
            sigsq=laobj.power_hmatrix(sigma,0.5)
            flag=1
        if flag==0:
            fidelity=laobj.trace_norm_rmatrix(np.matmul(rhosq,sigsq))
            fidelity=fidelity**2
        else:
            fidelity=laobj.trace_norm_cmatrix(np.matmul(rhosq,sigsq))
            fidelity=fidelity**2
        return fidelity
        
    def fidelity_vec2(self,vecrho,vecsigma):
        """
        Calculates fidelity between two quantum states
        Input:
            vecrho: input pure state vector.
            vecsigma: input pure state vector.
        Output:
            fidelity: it stores the value of fidelity
        """
        typerho=str(vecrho.dtype)
        if re.findall('^complex',typerho):
            fidelity=np.matmul(np.matrix.\
                               conjugate(np.matrix.transpose(vecrho)),\
                               vecsigma)
        else:
            fidelity=np.matmul(np.matrix.transpose(vecrho), vecsigma)
        fidelity=abs(fidelity)**2
        return fidelity
    
    def fidelity_vecden(self,vec,sigma):
        """
        Calculates fidelity between a quantum state and a density matrix
        Input:
            vec: input pure state vector.
            sigma: input density matrix
        Output:
            fidelity: it stores the value of fidelity
        """
        typevec=str(vec.dtype)
        if re.findall('^complex',typevec):
            fidelity=np.matmul(np.matrix.\
                               conjugate(np.matrix.transpose(vec)),\
                               np.matmul(sigma,vec))
        else:
            fidelity=np.matmul(np.matrix.transpose(vec),\
                               np.matmul(sigma,vec))
        fidelity=abs(fidelity)
        return fidelity
        
    def super_fidelity(self,rho,sigma):
        """
        Calculates super fidelity between two density matrices
        Input:
            rho: input density matrix.
            sigma: input density matrix.
        output:
            sf: the value of the super fidelity
        """
        tr_rho2=np.trace(np.matmul(rho,rho))
        tr_sigma2=np.trace(np.matmul(sigma,sigma))
        tr_rhosigma=np.trace(np.matmul(rho,sigma))
        sf=tr_rhosigma+np.sqrt((1-tr_rho2)*(1-tr_sigma2))
        sf=abs(sf)        
        return sf
        
    def bures_distance_vec(self,rho,sigma):
        """
        Calculates Bures distance between two quantum state
        Input:
            rho: input state vector
            sigma: input state vector
        Output:
            bd: the value of the Bures distance
        """
        fid=self.fidelity_vec2(rho,sigma)
        bd=np.sqrt(2*(1-np.sqrt(fid)))
        
        return bd
    
    def bures_distance_den(self,rho,sigma):
        """
        Calculates Bures distance between two density matrix
        Input:
            rho: input density matrix
            sigma: input density matrix
        Output:
            bd: the value of the Bures distance
        """
        fid=self.fidelity_den2(rho,sigma)
        bd=np.sqrt(2*(1-np.sqrt(fid)))
        
        return bd
    
    def expectation_vec(self,vec,obs):
        """
        Expectation values of observable for a quantum state
        Input:
            vec: input state vector
            obs: observable operator
        Output:
            expc: the expectation value of the measurement operator
        """
        typevec=str(vec.dtype)
        if re.findall('^complex',typevec):
            expc=np.matmul(np.matmul(np.matrix.\
                               conjugate(np.matrix.transpose(vec)),\
                               obs),vec)
        else:
            expc=np.matmul(np.matmul(np.matrix.transpose(vec),obs),vec)
        return expc
        
    def expectation_den(self,rho,obs):
        """
        Expectation values of observable for a density matrix
        Input:
            rho: input density matrix
            obs: observable operator
        Output:
            expc: the expectation value of the observable operator
        """
        return np.trace(np.matmul(rho,obs))
    

        
    

        
        
        
        
        


"""
Created on Mon Jul 15 18:26:57 2019

@author: Pranay Barkataki and Payal Dineshbhai Solanki
"""

def recur_comb_add(mylist,vec,icount,sum2):
    lenvec=len(vec)
    if icount<=lenvec-1:
        for j in range(icount,lenvec):
            sum3=sum2+vec[j]
            sum3=int(sum3)
            mylist.append(sum3)
            recur_comb_add(mylist,vec,j+1,sum3)
            if j==lenvec:
                return()
    if icount==lenvec:
        return()

"""
Created on Sun Aug  4 21:11:55 2019

@authors: Pranay Barkataki
"""
def RecurChainRL1(row,tot_spins,icount,mylist,shift):
    len_row=len(row)
    if icount<len_row:
        for x in range(icount,len_row):
            row2=row.copy()
            if x>=0:
                row2[x]=1
            if shift==0:
                y=0
            if shift==1:
                y=1
            sumr=0
            if shift==0:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0:
                        sumr=sumr+(2**y)
                    if row2[x1]==1:
                        sumr=sumr+(2**(y+1))
                    y=y+2
                mylist.append(sumr)
            if shift==1:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0 and x1==len_row-1:
                        sumr=sumr+(2**(tot_spins-1))
                    if row2[x1]==1 and x1==len_row-1:
                        sumr=sumr+1
                    if row2[x1]==0 and x1!=len_row-1:
                        sumr=sumr+(2**y)
                        y=y+2
                    if row2[x1]==1 and x1!=len_row-1:
                        sumr=sumr+(2**(y+1))
                        y=y+2
                mylist.append(sumr)
            if x<len_row-1 and x!=-1:    
                RecurChainRL1(row2,tot_spins,x+1,mylist,shift)
            if x==len_row-1:
                return()
    if icount >= len_row:
        return()

"""
Created on Sun Aug  7 14:19:31 2019

@authors: Pranay Barkataki
"""
def RecurChainRL2(row,tot_spins,icount,mylist,shift):
    len_row=len(row)
    if icount<len_row:
        for x in range(icount,len_row):
            row2=row.copy()
            if x>=0:
                row2[x]=1
            if shift==0:
                y=0
            if shift==1:
                y=1
            sumr=0
            cntr=0
            if shift==0:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0:
                        sumr=sumr+(2**y)
                        cntr=cntr+0
                    if row2[x1]==1:
                        sumr=sumr+(2**(y+1))
                        cntr=cntr+1
                    y=y+2
                if cntr%2==0:
                    mylist.append(sumr)
                if cntr%2==1:
                    mylist.append(-sumr)
            if shift==1:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0 and x1==len_row-1:
                        sumr=sumr+(2**(tot_spins-1))
                    if row2[x1]==1 and x1==len_row-1:
                        sumr=sumr+1
                        cntr=cntr+1
                    if row2[x1]==0 and x1!=len_row-1:
                        sumr=sumr+(2**y)
                        y=y+2
                    if row2[x1]==1 and x1!=len_row-1:
                        sumr=sumr+(2**(y+1))
                        y=y+2
                        cntr=cntr+1
                if cntr%2==0:
                    mylist.append(sumr)
                    #print(sumr)
                if cntr%2!=0:
                    mylist.append(-sumr)
                    #print(sumr)
            if x<len_row-1 and x!=-1:    
                RecurChainRL2(row2,tot_spins,x+1,mylist,shift)
            if x==len_row-1:
                return()
    if icount >= len_row:
        return()
        
"""
Created on Sun Aug  7 23:09:23 2019

@authors: Pranay Barkataki
"""
def RecurChainRL3(row,tot_spins,icount,mylist,shift):
    len_row=len(row)
    if icount<len_row:
        for x in range(icount,len_row):
            row2=row.copy()
            if x>=0:
                row2[x]=1
            if shift==0:
                y=0
            if shift==1:
                y=1
            sumr=0
            if shift==0:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0:
                        sumr=sumr+0
                    if row2[x1]==1:
                        sumr=sumr+(2**y)+(2**(y+1))
                    y=y+2
                mylist.append(sumr)
            if shift==1:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0 and x1==len_row-1:
                        sumr=sumr+0
                    if row2[x1]==1 and x1==len_row-1:
                        sumr=sumr+1+(2**(tot_spins-1))
                    if row2[x1]==0 and x1!=len_row-1:
                        sumr=sumr+0
                        y=y+2
                    if row2[x1]==1 and x1!=len_row-1:
                        sumr=sumr+(2**(y+1))+(2**y)
                        y=y+2
                mylist.append(sumr)
            if x<len_row-1 and x!=-1:    
                RecurChainRL3(row2,tot_spins,x+1,mylist,shift)
            if x==len_row-1:
                return()
    if icount >= len_row:
        return()

"""
Created on Sun Aug  7 14:19:31 2019

@authors: Pranay Barkataki
"""
def RecurChainRL4(row,tot_spins,icount,mylist,shift):
    len_row=len(row)
    if icount<len_row:
        for x in range(icount,len_row):
            row2=row.copy()
            if x>=0:
                row2[x]=1
            if shift==0:
                y=0
            if shift==1:
                y=1
            sumr=0
            cntr=0
            if shift==0:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0:
                        sumr=sumr+0
                        cntr=cntr+0
                    if row2[x1]==1:
                        sumr=sumr+(2**y)+(2**(y+1))
                        cntr=cntr+1
                    y=y+2
                if cntr%2==0:
                    mylist.append(sumr)
                if cntr%2==1:
                    mylist.append(-sumr)
            if shift==1:
                for x1 in range(len_row-1,-1,-1):
                    if row2[x1]==0 and x1==len_row-1:
                        sumr=sumr+0
                    if row2[x1]==1 and x1==len_row-1:
                        sumr=sumr+1+(2**(tot_spins-1))
                        cntr=cntr+1
                    if row2[x1]==0 and x1!=len_row-1:
                        sumr=sumr+0
                        y=y+2
                    if row2[x1]==1 and x1!=len_row-1:
                        sumr=sumr+(2**(y+1))+(2**y)
                        y=y+2
                        cntr=cntr+1
                if cntr%2==0:
                    mylist.append(sumr)
                    #print(sumr)
                if cntr%2!=0:
                    mylist.append(-sumr)
                    #print(sumr)
            if x<len_row-1 and x!=-1:    
                RecurChainRL4(row2,tot_spins,x+1,mylist,shift)
            if x==len_row-1:
                return()
    if icount >= len_row:
        return()


from .chap2_nutsboltsquantum import QuantumMechanics
from .chap3_linearalgebra import LinearAlgebra
from .chap4_quantumtools import GatesTools
from .chap5_partial_quantumentanglement import PartialTr
from .chap5_partial_quantumentanglement import Entanglement
from .chap6_hamiltonians import Hamiltonian
from .chap7_randommatrix import QRandom


"""
Created on Thu Aug 22 19:18:53 2019

@author: Pranay Barkataki 
"""

import numpy as np
import math
from QuantumInformation import RecurNum
from QuantumInformation import LinearAlgebra as LA
from QuantumInformation import QuantumMechanics as QM
import scipy.linalg.lapack as la
import re

qobj=QM()

class PartialTr:
    def __init__(self):
        """It is a class dealing with partial trace and transpose
        it primarily intrinsic functions of uses numpy, math, cmath.
        """
    # partial trace operation subroutine for a real pure state 
    #entry is in the column form
    def partial_trace_vec(self,state,sub_tr):
        """
        Partial trace operation on a quantum state
        Input:
            state: real state vector
            sub_tr: details of the subsystems not to be traced out
        Output:
            red_den: reduced density matrix
        """
        typestate=str(state.dtype)
        N=int(math.log2(state.shape[0]))
        length=len(sub_tr)
        # count=length, and count0= N-length
        assert set(sub_tr).issubset(set(np.arange(1,N+1))),\
        "Invalid subsystems to be traced out"
        if re.findall("^complex",typestate):
            red_den=np.zeros([2**(length),2**(length)],dtype=np.complex_)
            vec=np.zeros([(N-length),1])
            im=0
            for ii in range(1,N+1):
                if ii not in sub_tr:
                    vec[im]=2**(N-ii)
                    im=im+1
            mylist=[]
            icount=0
            sum2=0
            RecurNum.recur_comb_add(mylist,vec,icount,sum2)
            irow=np.zeros([N,1])
            icol=np.zeros([N,1])
            mylist=np.array(mylist)
            len_mylist=len(mylist)    
            for i1 in range(0,2**length):
                col1=self.__dectobin(i1,length)
                for i2 in range(0,2**length):
                    col2=self.__dectobin(i2,length)
                    i3=0
                    for k in range(0,N):
                        if k+1 not in sub_tr:
                            irow[k]=0
                        else:
                            irow[k]=col1[i3]
                            i3=i3+1 
                    ic=0
                    for k2 in range(0,N):
                        if k2+1 not in sub_tr:
                            icol[k2]=0
                        else:
                            icol[k2]=col2[ic]
                            ic=ic+1
                    icc=self.__bintodec(irow)
                    jcc=self.__bintodec(icol)
                    red_den[i1,i2]=red_den[i1,i2]+(state[icc]*\
                           np.conjugate(state[jcc]))
                    for jj in range(0,len_mylist):
                        icc2=icc+mylist[jj]
                        jcc2=jcc+mylist[jj]
                        red_den[i1,i2]=red_den[i1,i2]+(state[icc2]*\
                               np.conjugate(state[jcc2]))
        else:
            red_den=np.zeros([2**(length),2**(length)],dtype='float64')
            vec=np.zeros([(N-length),1])
            im=0
            for ii in range(1,N+1):
                if ii not in sub_tr:
                    vec[im]=2**(N-ii)
                    im=im+1
            mylist=[]
            icount=0
            sum2=0
            RecurNum.recur_comb_add(mylist,vec,icount,sum2)
            irow=np.zeros([N,1])
            icol=np.zeros([N,1])
            mylist=np.array(mylist)
            len_mylist=len(mylist)    
            for i1 in range(0,2**length):
                col1=self.__dectobin(i1,length)
                for i2 in range(0,2**length):
                    col2=self.__dectobin(i2,length)
                    i3=0
                    for k in range(0,N):
                        if k+1 not in sub_tr:
                            irow[k]=0
                        else:
                            irow[k]=col1[i3]
                            i3=i3+1 
                    ic=0
                    for k2 in range(0,N):
                        if k2+1 not in sub_tr:
                            icol[k2]=0
                        else:
                            icol[k2]=col2[ic]
                            ic=ic+1
                    icc=self.__bintodec(irow)
                    jcc=self.__bintodec(icol)
                    red_den[i1,i2]=red_den[i1,i2]+(state[icc]*state[jcc])
                    for jj in range(0,len_mylist):
                        icc2=icc+mylist[jj]
                        jcc2=jcc+mylist[jj]
                        red_den[i1,i2]=red_den[i1,i2]+(state[icc2]*state[jcc2])                   
        return(red_den)
    
    # partial trace operation for a real state density matrix    
    def partial_trace_den(self,state,sub_tr):
        """
        Partial trace operation on a density matrix
        Input:
            state: input real density matrix
            sub_tr: details of the subsystem not to be traced out
        Output:
            red_den: reduced density matrix
        """
        typestate=str(state.dtype)
        N=int(math.log2(state.shape[0]))
        length=len(sub_tr)
        # count=length, and count0= N-length
        assert set(sub_tr).issubset(set(np.arange(1,N+1))),\
        "Invalid subsystems to be traced out"
        if re.findall("^complex",typestate):
            red_den=np.zeros([2**(length),2**(length)],dtype=np.complex_)
            vec=np.zeros([(N-length),1])
            im=0
            for ii in range(1,N+1):
                if ii not in sub_tr:
                    vec[im]=2**(N-ii)
                    im=im+1
            mylist=[]
            icount=0
            sum2=0
            RecurNum.recur_comb_add(mylist,vec,icount,sum2)
            irow=np.zeros([N,1])
            icol=np.zeros([N,1])
            mylist=np.array(mylist)
            len_mylist=len(mylist)    
            for i1 in range(0,2**length):
                col1=self.__dectobin(i1,length)
                for i2 in range(0,2**length):
                    col2=self.__dectobin(i2,length)
                    i3=0
                    for k in range(0,N):
                        if k+1 not in sub_tr:
                            irow[k]=0
                        else:
                            irow[k]=col1[i3]
                            i3=i3+1 
                    ic=0
                    for k2 in range(0,N):
                        if k2+1 not in sub_tr:
                            icol[k2]=0
                        else:
                            icol[k2]=col2[ic]
                            ic=ic+1
                    icc=self.__bintodec(irow)
                    jcc=self.__bintodec(icol)
                    red_den[i1,i2]=red_den[i1,i2]+(state[icc,jcc])
                    for jj in range(0,len_mylist):
                        icc2=icc+mylist[jj]
                        jcc2=jcc+mylist[jj]
                        red_den[i1,i2]=red_den[i1,i2]+(state[icc2,jcc2])
        else:
            red_den=np.zeros([2**(length),2**(length)],dtype='float64')
            vec=np.zeros([(N-length),1])
            im=0
            for ii in range(1,N+1):
                if ii not in sub_tr:
                    vec[im]=2**(N-ii)
                    im=im+1
            mylist=[]
            icount=0
            sum2=0
            RecurNum.recur_comb_add(mylist,vec,icount,sum2)
            irow=np.zeros([N,1])
            icol=np.zeros([N,1])
            mylist=np.array(mylist)
            len_mylist=len(mylist)    
            for i1 in range(0,2**length):
                col1=self.__dectobin(i1,length)
                for i2 in range(0,2**length):
                    col2=self.__dectobin(i2,length)
                    i3=0
                    for k in range(0,N):
                        if k+1 not in sub_tr:
                            irow[k]=0
                        else:
                            irow[k]=col1[i3]
                            i3=i3+1 
                    ic=0
                    for k2 in range(0,N):
                        if k2+1 not in sub_tr:
                            icol[k2]=0
                        else:
                            icol[k2]=col2[ic]
                            ic=ic+1
                    icc=self.__bintodec(irow)
                    jcc=self.__bintodec(icol)
                    red_den[i1,i2]=red_den[i1,i2]+(state[icc,jcc])
                    for jj in range(0,len_mylist):
                        icc2=icc+mylist[jj]
                        jcc2=jcc+mylist[jj]
                        red_den[i1,i2]=red_den[i1,i2]+(state[icc2,jcc2])                   
        return(red_den)
        
    # Partial Transpose of real pure state
    def ptranspose_vec(self,state,sub_tr):
        """
        Partial transpose operation on a quantum state
        Parameters
            state : It is a real or complex state.
            sub_tr : List of number designating the subsystems
                     to be partially transposed.
        Returns
            denc2: It is partially transposed density matrix

        """
        N=int(math.log2(state.shape[0]))
        assert set(sub_tr).issubset(set(np.arange(1,N+1))),\
        "Invalid subsystems to be traced out"
        typestate=str(state.dtype)
        if re.findall("^complex",typestate):
            denc2=np.zeros([2**N,2**N],dtype=np.complex_)
            for i in range(state.shape[0]):
                vec_row=qobj.decimal_binary(i,N)
                for j in range(state.shape[0]):
                    vec_col=qobj.decimal_binary(j,N)
                    vec_row2=vec_row.copy()
                    for k in sub_tr:
                        temp=vec_row2[k-1]
                        vec_row2[k-1]=vec_col[k-1]
                        vec_col[k-1]=temp
                    row=qobj.binary_decimal(vec_row2)
                    col=qobj.binary_decimal(vec_col)
                    denc2[row,col]=state[i]*np.conjugate(state[j])
        else:            
            denc2=np.zeros([2**N,2**N],dtype='float64')
            for i in range(state.shape[0]):
                vec_row=qobj.decimal_binary(i,N)
                for j in range(state.shape[0]):
                    vec_col=qobj.decimal_binary(j,N)
                    vec_row2=vec_row.copy()
                    for k in sub_tr:
                        temp=vec_row2[k-1]
                        vec_row2[k-1]=vec_col[k-1]
                        vec_col[k-1]=temp
                    row=qobj.binary_decimal(vec_row2)
                    col=qobj.binary_decimal(vec_col)
                    denc2[row,col]=state[i]*state[j]
        return(denc2)
    
    # Partial Transpose of real density matrix
    def ptranspose_den(self,denc,sub_tr):
        """
        Partial transpose operation on density matrix
        Parameters
            denc : It is a real or complex density matrix.
            sub_tr : List of number designating the subsystems
                     to be partially transposed.
        Returns
            denc2: It is partially transposed density matrix
        """
        N=int(math.log2(denc.shape[0]))
        assert set(sub_tr).issubset(set(np.arange(1,N+1))),\
        "Invalid subsystems to be traced out"
        typestate=str(denc.dtype)
        if re.findall("^complex",typestate):
            denc2=np.zeros([2**N,2**N],dtype=np.complex_)
            for i in range(denc.shape[0]):
                vec_row=qobj.decimal_binary(i,N)
                for j in range(denc.shape[1]):
                    vec_col=qobj.decimal_binary(j,N)
                    vec_row2=vec_row.copy()
                    for k in sub_tr:
                        temp=vec_row2[k-1]
                        vec_row2[k-1]=vec_col[k-1]
                        vec_col[k-1]=temp
                    row=qobj.binary_decimal(vec_row2)
                    col=qobj.binary_decimal(vec_col)
                    denc2[row,col]=denc[i,j]
        else:
            denc2=np.zeros([2**N,2**N],dtype='float64')
            for i in range(denc.shape[0]):
                vec_row=qobj.decimal_binary(i,N)
                for j in range(denc.shape[1]):
                    vec_col=qobj.decimal_binary(j,N)
                    vec_row2=vec_row.copy()
                    for k in sub_tr:
                        temp=vec_row2[k-1]
                        vec_row2[k-1]=vec_col[k-1]
                        vec_col[k-1]=temp
                    row=qobj.binary_decimal(vec_row2)
                    col=qobj.binary_decimal(vec_col)
                    denc2[row,col]=denc[i,j]
        return(denc2)
        
    def __dectobin(self,n,l):
        """It converts decimal to binary.
        Attributes:
            n: entry of the decimal number
            l: length of the binary output
        Returns:
            dtb: a numpy array containing the binary equivalent of number n
        """
        import numpy as np
        p=n
        dtb=np.empty([l,1])
        for i in range(0,l):
            dtb[l-1-i]=int(p % 2)
            p=int(p/2)
        #print(dtb)    
        return(dtb)
        
    # Binary to decimal conversion
    def __bintodec(self,vec):
        """
        It converts biinary to decimal
        Attributes:
            vec: entry of 1D array of binary numbers {0,1}
        Returns:
            t: decimal equivalent of the vec
        """
        t=0
        for i in range(0,len(vec)):
            t=t+vec[len(vec)-1-i]*(2**i)
        #print(dtb)    
        return(int(t))


class Entanglement(PartialTr):
    
    # Concurrence calculation for a real pure state
    def concurrence_vec(self,state,i,j,eps=10**(-13)):
        """
        Calculation of concurrence for a quantum state
        Parameters
            state : Real or complex state
            i : It stores the place values of the qubits.
            j : It stores the place values of the qubits.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            conc: concurrence value
        """
        sigmay=np.zeros([4,4],dtype='float64')
        typestate=str(state.dtype)
        if re.findall("^complex",typestate):
            sigmay[0,3]=-1
            sigmay[1,2]=1
            sigmay[2,1]=1
            sigmay[3,0]=-1
            sub_tr=[i,j]
            rdm= self.partial_trace_vec(state,sub_tr)
            rhot3=rdm@sigmay@np.conjugate(rdm)@sigmay
            w,vl,vr,info =la.zgeev(rhot3)
            wc=[]
            for i in range(0,4):
                if abs(w.item(i))<eps:
                    wc.append(0.000000000000000)
                else:
                    wc.append(abs(w.item(i)))
            wc.sort(reverse=True)
            wc=np.array(wc,dtype='float64')
            conc=math.sqrt(wc.item(0))-math.sqrt(wc.item(1))-\
            math.sqrt(wc.item(2))-math.sqrt(wc.item(3))
            if conc<0:
                conc=0
        else:    
            sigmay[0,3]=-1
            sigmay[1,2]=1
            sigmay[2,1]=1
            sigmay[3,0]=-1
            sub_tr=[i,j]
            rdm= self.partial_trace_vec(state,sub_tr)
            rhot3=rdm@sigmay@rdm@sigmay
            wr,wi,vl,vr,info =la.dgeev(rhot3)
            w=[]
            for i in range(0,4):
                if wr[i] < eps:
                    w.append(0.000000000000000)
                else:
                    w.append(np.float64(wr.item(i)))
            w.sort(reverse=True)
            w=np.array(w,dtype='float64')
            conc=math.sqrt(w.item(0))-math.sqrt(w.item(1))-\
            math.sqrt(w.item(2))-math.sqrt(w.item(3))
            if conc<0:
                conc=0.0
        return(np.float64(conc))
    
    # Concurrence calculation for real state density matrix 
    def concurrence_den(self,state,i,j,eps=10**(-13)):
        """
        Calculation of concurrence for a density matrix
        Parameters
            state : Real or complex density matrix
            i : It stores the place values of the qubits.
            j : It stores the place values of the qubits.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            conc: concurrence value
        """
        sigmay=np.zeros([4,4],dtype='float64')
        typestate=str(state.dtype)
        if re.findall("^complex",typestate):
            sigmay[0,3]=-1
            sigmay[1,2]=1
            sigmay[2,1]=1
            sigmay[3,0]=-1
            sub_tr=[i,j]
            rdm= self.partial_trace_den(state,sub_tr)
            rhot3=rdm@sigmay@np.conjugate(rdm)@sigmay
            w,vl,vr,info =la.zgeev(rhot3)
            wc=[]
            for i in range(0,4):
                if abs(w.item(i))<eps:
                    wc.append(0.000000000000000)
                else:
                    wc.append(abs(w.item(i)))
            wc.sort(reverse=True)
            wc=np.array(wc,dtype='float64')
            conc=math.sqrt(wc.item(0))-math.sqrt(wc.item(1))-\
            math.sqrt(wc.item(2))-math.sqrt(wc.item(3))
            if conc<0:
                conc=0
        else:
            sigmay[0,3]=-1
            sigmay[1,2]=1
            sigmay[2,1]=1
            sigmay[3,0]=-1
            sub_tr=[i,j]
            rdm= self.partial_trace_den(state,sub_tr)
            rhot3=rdm@sigmay@rdm@sigmay
            wr,wi,vl,vr,info =la.dgeev(rhot3)
            w=[]
            for i in range(0,4):
                if wr[i] < eps:
                    w.append(0.000000000000000)
                else:
                    w.append(np.float64(wr.item(i)))
            w.sort(reverse=True)
            w=np.array(w,dtype='float64')
            conc=math.sqrt(w.item(0))-math.sqrt(w.item(1))-\
            math.sqrt(w.item(2))-math.sqrt(w.item(3))
            if conc<0:
                conc=0.0
        return(np.float64(conc))    
    
    # Block entropy for a pure real state
    def block_entropy_vec(self,state,sub_tr,eps=10**(-13)):
        """
        Calculation of block entropy for a quantum state
        Parameters
            state : Real or complex state
            sub_tr: List of numbers designating the particular subsystems
                    not to be traced out.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            Bent: Block entropy value
        """
        typestate=str(state.dtype)
        rdm= self.partial_trace_vec(state,sub_tr)
        if re.findall("^complex",typestate):
            w,v,info=la.zheev(rdm)
        else:
            w,v,info=la.dsyev(rdm)
        wlen=len(w)
        Bent=0.0
        for x in range(0,wlen):
            if abs(w.item(x))<eps:
                w[x]=0.000000000000000
            else:
                assert w.item(x) > 0.0,\
                "The density matrix entered is not correct as the eigenvalues are negative"
                Bent=Bent-(w.item(x)*math.log(w.item(x),2))
        return(Bent)

    # Block entropy for a pure real density matrix
    def block_entropy_den(self,state,sub_tr,eps=10**(-13)):
        """
        Calculation of block entropy for a density matrix
        Parameters
            state : Real or complex density matrix
            sub_tr: List of numbers designating the particular subsystems
                    not to be traced out.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            Bent: Block entropy value
        """
        typestate=str(state.dtype)
        rdm= self.partial_trace_den(state,sub_tr)
        if re.findall("^complex",typestate):
            w,v,info=la.zheev(rdm)
        else:
            w,v,info=la.dsyev(rdm)
        wlen=len(w)
        Bent=0.0
        for x in range(0,wlen):
            if abs(w.item(x))<eps:
                w[x]=0.000000000000000
            else:
                assert w.item(x) > 0.0,\
                "The density matrix entered is not correct as the eigenvalues are negative"
                Bent=Bent-(w.item(x)*math.log(w.item(x),2))
        return(Bent)
       
    # Q measure for pure real state
    def QMeasure_vec(self,state): 
        """
        Calculation of Q measure for a quantum state
        Parameters
            state : Real or complex state

        Returns
            Qmeas: Q measure value

        """
        NN=math.log2(state.shape[0])/math.log2(2)
        NN=int(NN)
        sub_tr=np.zeros([NN,1])
        sum3=0.0
        for x in range(0,NN):
            sub_tr=[]
            sub_tr.append(x+1)
            rho=self.partial_trace_vec(state,sub_tr)
            rho=np.matmul(rho,rho)
            tr2=np.trace(rho)
            sum3=sum3+tr2
        Qmeas=2*(1-(sum3/NN))
        return abs(Qmeas)
        
    # Q measure for real density matrix
    def QMeasure_den(self,den):
        """
        Calculation of Q measure for a density matrix
        Parameters
            den : Real or complex density matrix

        Returns
            Qmeas: Q measure value

        """
        NN=math.log2(den.shape[0])/math.log2(2)  
        NN=int(NN)
        sub_tr=np.zeros([NN,1])
        sum3=0.0
        for x in range(0,NN):
            sub_tr=[]
            sub_tr.append(x+1)
            rho=self.partial_trace_den(den,sub_tr)
            rho=np.matmul(rho,rho)
            tr2=np.trace(rho)
            sum3=sum3+tr2
        Qmeas=2*(1-(sum3/NN))
        return abs(Qmeas)

    # Negativity of real pure state
    def negativity_log_vec(self,state,sub_tr,eps=10**(-13)):
        """
        Calculation of negativity and logarithmic negativity for a quantum state
        Parameters
            state : Real or complex state
            sub_tr: List of numbers designating the particular subsystems
                    to be transposed.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            negv,lognegv : negativity and log negativity values, respectively
        """
        laobj=LA()
        typestate=str(state.dtype)
        rhoa=self.ptranspose_vec(state,sub_tr)
        if re.findall("^complex",typestate):
            negv=laobj.trace_norm_cmatrix(rhoa,precision=eps)
        else:
            negv=laobj.trace_norm_rmatrix(rhoa,precision=eps)
        assert negv > 0.0,\
        "The density matrix entered is not correct as the negativity is negative"
        lognegv=math.log2(negv)
        negv=(negv-1)/2
        return(negv,lognegv)
    
    # Negativity of real pure state
    def negativity_log_den(self,den,sub_tr,eps=10**(-13)):
        """
        Calculation of negativity and logarithmic negativity for a density matrix
        Parameters
            state : Real or complex density matrix
            sub_tr: List of numbers designating the particular subsystems
                    to be transposed.
            eps : Below the eps value the eigenvalues will be considered zero.
                  The default is 10**(-13).

        Returns
            negv,lognegv : negativity and log negativity values, respectively
        """
        laobj=LA()
        typestate=str(den.dtype)
        rhoa=self.ptranspose_den(den,sub_tr)
        if re.findall("^complex",typestate):
            negv=laobj.trace_norm_cmatrix(rhoa,precision=eps)
        else:
            negv=laobj.trace_norm_rmatrix(rhoa,precision=eps)
        assert negv > 0.0,\
        "The density matrix entered is not correct as the negativity is negative"
        lognegv=math.log2(negv)
        negv=(negv-1)/2
        return(negv,lognegv)
        
    def renyi_entropy(self,rho,alpha):
        """
        Calculation of Renyi entropy
        Parameters
            rho : Real or complex density matrix
            alpha : It is the value of Renyi index

        Returns
            renyi : Renyi Entropy value

        """
        assert alpha != 1.0, "alpha should not be equal to 1"
        typerho=str(rho.dtype)
        laobj=LA()
        if re.findall('^complex',typerho):
            renyi=math.log(abs(np.trace(laobj.power_hmatrix(rho,alpha))))/(1-alpha)
        else:
            renyi=math.log(np.trace(laobj.power_smatrix(rho,alpha)))/(1-alpha)
        return renyi
    
    def entanglement_spectrum(self,rho):
        """
        Calculation of entanglement spectrum of a density matrix
        Parameters
            rho :  Real or complex density matrix

        Returns
            eigenvalues : List containing the eigenvalues of rho
            logeigenvalues : List containing the negative logarithmic
                            eigenvalues of rho
        """
        typerho=str(rho.dtype)
        if re.findall('^complex',typerho):
            eigenvalues,eigenvectors,info=la.zheev(rho)
        else:
            eigenvalues,eigenvectors,info=la.dsyev(rho)
        logeigenvalues=np.zeros([eigenvalues.shape[0]],dtype='float64')
        for i in range(0,eigenvalues.shape[0]):
            assert eigenvalues[i]>0.0,\
            "The eigenvalues of the matrix is coming less than equal to zero"
            logeigenvalues[i]=(-1)*math.log(eigenvalues[i])
        return (eigenvalues,logeigenvalues)
    
    def residual_entanglement_vec(self,state):
        """
        Calculation of residual entanglement for a three-qubit quantum state
        Parameters
            state : Real or complex 3-qubit state

        Returns
            res_tang : Residual entanglement value

        """
        assert state.shape[0]==8,"It is not a three qubit quantum system"
        det=np.linalg.det(self.partial_trace_vec(state,[1]))
        det=4*det
        res_tang=det-(self.concurrence_vec(state,1,2)**2)-\
        (self.concurrence_vec(state,1,3)**2)
        res_tang=abs(res_tang)
        return res_tang
    
    def residual_entanglement_den(self,den):
        """
        Calculation of residual entanglement for a three-qubit density matrix
        Parameters
            den : Real or complex 3-qubit density matrix

        Returns
            res_tang : Residual entanglement value
        """
        assert den.shape[0]==8,"It is not a three qubit quantum system"
        det=np.linalg.det(self.partial_trace_den(den,[1]))
        det=4*det
        res_tang=det-(self.concurrence_den(den,1,2)**2)-\
        (self.concurrence_den(den,1,3)**2)
        res_tang=abs(res_tang)
        return res_tang







"""
Created on Thu Aug 22 19:18:53 2019

@author: Pranay Barkataki 
"""
#import scipy.linalg.lapack as la
import numpy as np
import re

class QRandom:
    def __init__(self):
        """It is a class dealing with partial trace and transpose
        it primarily intrinsi functions of uses numpy, math, cmath.
        """
        
    def random_gaussian_rvec(self,tup,mu=0,sigma=1):
        """
        Construct a real random state from a Gaussian distribution 
        Input:
            tup: tuple holds the dimension of the output matrix.
            mu: it is the average value of the gaussian distribution.
            sigma: it is the standard deviation of the gaussian distribution
        Output:
            gauss_state: it the gaussian distributed real state
        """
        gauss_state=np.random.normal(mu, sigma, tup)
        return gauss_state
    
    def random_gaussian_cvec(self,tup,mu=0,sigma=1):
        """
        Construct a complex random state from a Gaussian distribution
        Input:
            tup: tuple holds the dimension of the output matrix.
            mu: it is the average value of the gaussian distribution.
            sigma: it is the standard deviation of the gaussian distribution
        Output:
            gauss_state: it the gaussian distributed complex state
        """
        rpart=np.random.normal(mu, sigma, tup)
        cpart=np.random.normal(mu, sigma, tup)
        state_gauss=np.zeros(tup,dtype=np.complex_)
        state_gauss=rpart+(complex(0,1)*cpart)
        return state_gauss
    
    def random_unifrom_rvec(self,tup,low=0.0,high=1.0):
        """
        Construct a real random state from a uniform distribution
        Input:
            tup: tuple holds the dimension of the output matrix.
            low: it is the lower bound of the uniform distribution.
            high: it is the upper bound of the uniform distribution
        Output:
            uniform_state: it the uniform distributed real state
        """
        uniform_state=np.random.uniform(low=low, high=high, size=tup)
        return uniform_state
       
    def random_unifrom_cvec(self,tup,low=0.0,high=1.0):
        """
        Construct a complex random state from a uniform distribution
        Input:
            tup: tuple holds the dimension of the output matrix.
            low: it is the lower bound of the uniform distribution.
            high: it is the upper bound of the uniform distribution
        Output:
            uniform_state: it the uniform distributed complex state
        """
        uniform_rpart=np.random.uniform(low=low, high=high, size=tup)
        uniform_cpart=np.random.uniform(low=low, high=high, size=tup)
        uniform_state=uniform_rpart+(complex(0,1)*uniform_cpart)
        return uniform_state
    
    def random_symmetric_matrix(self,size,distribution="gaussian",mu=0,\
                                sigma=1,low=0.0,high=1.0):
        """
        Construct a random symmetric matrix
        Input:
            size: it is a tuple (a,b), where a in number of rows and b is 
                  number of columns of the symmetric matrix
            distribution: it is the type of distribution
        Output:
            smatrix: symmetric matrix 
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        if distribution == 'gaussian':
            smatrix=np.random.normal(mu, sigma, size=size)
        if distribution == 'uniform':
            smatrix=np.random.uniform(low=low, high=high, size=size)
        smatrix=smatrix+np.matrix.transpose(smatrix)
        smatrix=0.5*smatrix
        return smatrix
    
    def random_hermitian_matrix(self,size,distribution="gaussian",mu=0,\
                                sigma=1,low=0.0,high=1.0):
        """
        Construct a random Hermitian matrix
        Input:
            size: it is a tuple (a,b), where a in number of rows and b is 
                  number of columns of the hermitian matrix
            distribution: it is the type of distribution
        Output:
            hmatrix: hermitian matrix 
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        hmatrix=np.zeros([size[0],size[1]],dtype=np.complex_)
        if distribution=='gaussian':
            for i in range(0,hmatrix.shape[0]):
                for j in range(0,hmatrix.shape[1]):
                    hmatrix[i,j]=complex(np.random.normal(mu, sigma, size= None),\
                           np.random.normal(mu, sigma, size= None))
        else:
            for i in range(0,hmatrix.shape[0]):
                for j in range(0,hmatrix.shape[1]):
                    hmatrix[i,j]=complex(np.random.uniform(low=low, high=high,\
                           size=None),np.random.uniform(low=low, high=high,\
                                    size=None))
        hmatrix=hmatrix+np.matrix.conjugate(np.matrix.transpose(hmatrix))
        hmatrix=0.5*hmatrix
        return hmatrix
    
    def random_orthogonal_matrix(self,size,distribution="gaussian",mu=0,\
                              sigma=1,low=0.0,high=1.0):
        """
        Construct a random orthogonal matrix
        Input:
            size: it is a tuple (a,b), where a in number of rows and b is 
                  number of columns of the orthogonal matrix
            distribution: it is the type of distribution
        Output:
            omatrix: orthogonal matrix 
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        
        if distribution == 'gaussian':
            omatrix=np.random.normal(mu, sigma, size= size)
        else:
            omatrix=np.random.uniform(low=low, high=high, size=size)
        omatrix,r=np.linalg.qr(omatrix,mode='complete')
        return omatrix
    
    def random_unitary_matrix(self,size,distribution="gaussian",mu=0,\
                              sigma=1,low=0.0,high=1.0):
        """
        Construct a random unitary matrix
        Input:
            size: it is a tuple (a,b), where a in number of rows and b is 
                  number of columns of the unitary matrix
            distribution: it is the type of distribution
        Output:
            umatrix: unitary matrix 
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        umatrix=np.zeros([size[0],size[1]],dtype=np.complex_)
        
        if distribution == 'gaussian':
            for i in range(0,umatrix.shape[0]):
                for j in range(0,umatrix.shape[1]):
                    umatrix[i,j]=complex(np.random.normal(mu,sigma,size=None),\
                           np.random.normal(mu,sigma,size=None))
        else:
            for i in range(0,umatrix.shape[0]):
                for j in range(0,umatrix.shape[1]):
                    umatrix[i,j]=complex(\
                           np.random.uniform(low=low,high=high,size=None),\
                           np.random.uniform(low=low,high=high,size=None))
        umatrix,r=np.linalg.qr(umatrix,mode='complete')
        return umatrix
    
    def random_real_ginibre(self,N):
        """
        Construct a real Ginibre matrix
        Input:
            N: dimension of the NxN Ginibre matrix
        Output:
            real ginibre matrix
        """
        return np.random.normal(0.0, 1.0, size= (N,N))
    
    def random_complex_ginibre(self,N):
        """
        Construct a complex Ginibre matrix
        Input:
            N: dimension of the NxN complex Ginibre
        Output:
            cginibre: complex ginibre matrix
        """
        cginibre=np.zeros([N,N],dtype=np.complex_)
        for i in range(0,cginibre.shape[0]):
            for j in range(0,cginibre.shape[1]):
                cginibre[i,j]=complex(np.random.normal(0.0, 1.0, size= None),\
                        np.random.normal(0.0, 1.0, size= None))
        return cginibre
    
    def random_real_wishart(self,N):
        """
        Construct a real Wishart matrix
        Input:
            N: dimension of the NxN real Wishart matrix
        Output:
            rwishart: real Wishart matrix
        """
        g=self.random_real_ginibre(N)
        rwishart=np.matmul(g,np.matrix.transpose(g))
        return rwishart
    
    def random_complex_wishart(self,N):
        """
        Construct a complex Wishart matrix
        Input:
            N: dimension of the NxN complex Wishart matrix
        Output:
            cwishart: complex Wishart matrix
        """
        g=self.random_complex_ginibre(N)
        cwishart=np.matmul(g,np.matrix.conjugate(np.matrix.transpose(g)))
        return cwishart
    
    def random_probability_vec(self,N):
        """
        Constructs a random probability vector
        Input:
            N: dimension of the probability vector.
        Output:
            prob_vec: The probability vector
        """
        prob_vec=np.random.uniform(low=0,high=1.0,size=N)
        norm=prob_vec.sum()
        prob_vec=prob_vec/norm
        return prob_vec
    
    def random_qrstate(self,N,distribution='gaussian',\
                       mu=0,sigma=1,low=0.0,high=1.0):
        """
        Constructs a random real pure quantum state
        Input:
            N: number of qubits
            distribution: it is the type of distribution
        Output:
            qrstate: real quantum state
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        if distribution=='gaussian':
            qrstate=self.random_gaussian_rvec(2**N,mu=0,sigma=1)
        else:
            qrstate=self.random_unifrom_rvec(2**N,low=0.0,high=1.0)
        norm=np.matmul(np.matrix.transpose(qrstate),qrstate)
        qrstate=qrstate/np.sqrt(norm)
        return qrstate
    
    def random_qcstate(self,N,distribution='gaussian',\
                       mu=0,sigma=1,low=0.0,high=1.0):
        """
        Constructs a random complex pure quantum state
        Input:
            N: number of qubits
            distribution: it is the type of distribution
        Output:
            qcstate: complex quantum state
        """
        assert re.findall("^gaussian|^uniform",distribution),\
        "Invalid distribution type"
        if distribution=='gaussian':
            qcstate=self.random_gaussian_cvec(2**N,mu=0,sigma=1)
        else:
            qcstate=self.random_unifrom_cvec(2**N,low=0.0,high=1.0)
        norm=abs(np.matmul(\
                       np.matrix.conjugate(\
                                           np.matrix.transpose(qcstate)),\
                                           qcstate))
        qcstate=qcstate/np.sqrt(norm)
        return qcstate
    
    def random_rden(self,N):
        """
        Constructs a random real pure density matrix
        Input:
            N: number of qubits
        Output:
            rden: real random density matrix
        """
        rden=self.random_real_wishart(2**N)
        rden=rden/np.trace(rden)
        return rden
    
    def random_cden(self,N):
        """
        Input:
            N: number of qubits
        Output:
            cden: complex random density matrix
        """
        cden=self.random_complex_wishart(2**N)
        cden=cden/np.trace(cden)
        return cden
        
    
        
        
        
            
        
        
        
        
        
        
        
        
        
        
        
        
        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan 23 2021

@author: pranay barkataki
"""

import numpy as np
import math

class QuantumMechanics:
    
    def __init__(self):
        """It is a quantum mechanics class, and it primarily intrinsic
            functions of uses numpy and scipy
        """
    
    def inner_product(self,vec1,vec2):
        """
        Here we compute the inner product 
        Attributes:
            vec1: it is a column vector.
            vec2: it is the second vector.
        Returns:
            inn: inner product between vec1 and vec2.
        """
        inn=complex(0.0,0.0)
        assert len(vec1) ==  len(vec2),\
        "Dimension of two vectors not equal to each other"
        for i in range(0,len(vec1)): 
            inn=inn+(np.conjugate(vec1[i])*vec2[i])
            
        return inn
    
    def norm_vec(self,vec):
        """
        Here we calculate norm of a vector
        Attributes:
            vec: column vector 
        Returns:
            norm: it contains the norm of the column vector vec
        """
        norm=0.0
        for i in range(0,len(vec)):
            norm=norm+abs(np.conjugate(vec[i])*vec[i])
        return np.sqrt(norm)
    
    def normalization_vec(self,vec):
        """
        Here normalize a given vector
        Attributes:
            vec: unnormalized column vector
        Returns:
            vec: normalized column vector
        """
        norm=0.0
        for i in range(0,len(vec)):
            norm=norm+abs(np.conjugate(vec[i])*vec[i])
        vec=vec/np.sqrt(norm)
        return vec
    
    def outer_product_rvec(self,vec1,vec2):
        """
        Here we calculate the outer product
        Attributes:
            vec1: it is a column real vector
            vec2: it is another real vector
        Returns:
            matrix: outer product of vec1 and vec2
        """
        matrix = np.zeros([len(vec1),len(vec2)],dtype='float64')
        for i in range(0,len(vec1)):
            for j in range(0,len(vec2)):
                matrix[i,j]=vec1[i]*vec2[j]
        return matrix
    
    def outer_product_cvec(self,vec1,vec2):
        """
        Here we calculate the outer product
        Attributes:
            vec1: it is a column complex vector
            vec2: it is another complex vector
        Returns:
            matrix: outer product of vec1 and vec2
        """
        matrix = np.zeros([len(vec1),len(vec2)],dtype=np.complex_)
        for i in range(0,len(vec1)):
            for j in range(0,len(vec2)):
                matrix[i,j]=vec1[i]*np.conjugate(vec2[j])
        return matrix    
    
    def tensor_product_matrix(self,A,B):
        """
        Here we calculate the tensor product of two matrix A and B
        Attributes:
            A: it is either a 1D or 2D array
            B: it is either a 1D or 2D array
        Returns: tensor product of A and B        
        """
        if len(A.shape)==1:
            A=A.reshape(A.shape[0],1)
        if len(B.shape)==1:
            B=B.reshape(B.shape[0],1)    
        return np.kron(A,B)                
    
    def commutation(self,A,B):
        """
        Here we calculate commutation between matrices A and B 
        Attributes:
            A: it is a square matrix
            B: it is another square matrix
        Returns: AB-BA matrix 
        """
        return np.matmul(A,B)-np.matmul(B,A)
    
    def anti_commutation(self,A,B):
        """
        Here we calculate anti commutation between matrices A and B 
        Attributes:
            A: input a square matrix
            B: input another square matrix
        Returns: AB+BA matrix
        """
        return np.matmul(A,B)+np.matmul(B,A)
    
    def rstate_shift(self,vec,shift=1,shift_direction='right'):
        """
        Shifting of a real quantum state
        Inputs:
            vec: real quantum state.
            shift: degree of the state.
            shift_direction: It shows the direction of the shift, 
                             by default it is right
        Return:
            state2: shifted state of vec
        """
        assert vec.shape[0]%2==0,"Not a qubit quantum state"
        N=int(math.log(vec.shape[0],2))
        basis=np.zeros([N],dtype=int)
        state2=np.zeros([2**N],dtype='float64')
        for i in range(0,vec.shape[0]):
            if vec[i] != 0.0:
                basis=self.decimal_binary(i,N)
                basis=self.binary_shift(basis,shift=shift,\
                                        shift_direction=shift_direction)
                j=int(self.binary_decimal(basis))
                state2[j]=vec[i]
        
        return state2
    
    def cstate_shift(self,vec,shift=1,shift_direction='right'):
        """
        Shifting of a complex quantum state
        Inputs:
            vec: complex quantum state.
            shift: degree of the state.
            shift_direction: It shows the direction of the shift, by default 
                            it is right
        Return:
            state2: shifted state of vec
        """
        assert vec.shape[0]%2==0,"Not a qubit quantum state"
        N=int(math.log(vec.shape[0],2))
        basis=np.zeros([N],dtype=int)
        state2=np.zeros([2**N],dtype=np.complex_)
        for i in range(0,vec.shape[0]):
            if vec[i] != complex(0.0,0.0):
                basis=self.decimal_binary(i,N)
                basis=self.binary_shift(basis,shift=shift,\
                                        shift_direction=shift_direction)
                j=self.binary_decimal(basis)
                state2[j]=vec[i]
        
        return state2
    
    def decimal_binary(self,i,N):
        """
        Decimal to binary conversion
        Inputs:
            i: input decimal number to changed to binary number
            N: number of the binary string
        Returns:
            bnum: binary number equivalent to i, it is column matrix.
        """
        bnum=np.zeros([N],dtype=int)
        for j in range(0,N):
            bnum[bnum.shape[0]-1-j]=i%2
            i=int(i/2)
        return bnum
            
    def binary_decimal(self,vec):
        """
        Binary to decimal conversion
        Input:
            vec:  array containing binary matrix elements i.e. 0 or 1
        Returns:
            dec: decimal equivalent number of binary array vec
        """
        dec=0
        for i in range(0,vec.shape[0]):
            dec=dec+int((2**(vec.shape[0]-1-i))*vec[i])
        dec=int(dec)
        return dec
    
    def binary_shift(self,vec,shift=1,shift_direction='right'):
        """
        Shifting of string of binary number
        Input:
            vec: array containing binary matrix elements i.e. 0 or 1
            shift: degree of the shift
            shift_direction: its value is either left or right
        Output:
            shift_vec: Shifted vec
        """
        assert shift_direction == 'left' or shift_direction == 'right',\
        'Not proper shift direction'
        assert shift >= 1 and shift < vec.shape[0],\
        "degree of shift is not proper"
        shift_vec=np.zeros([vec.shape[0]],dtype='int_')
        if shift_direction =='right':
            for i in range(0,vec.shape[0]):
                ishift=i+shift
                if ishift >= vec.shape[0]:
                    ishift=ishift-vec.shape[0]
                shift_vec[ishift]=vec[i]
        if shift_direction == 'left':
            for i in range(0,vec.shape[0]):
                ishift=i-shift
                if ishift < 0:
                    ishift=ishift+vec.shape[0]
                shift_vec[ishift]=vec[i]
        
        return shift_vec












    


"""
Created on Wed Nov  4 

@author: pranay barkataki
"""
import numpy as np
import math
import cmath
import scipy.linalg.lapack as la


class LinearAlgebra:
        
    def inverse_matrix(self,mat):
        """ Calculates the inverse of a matrix
            Attributes:
                mat : Inverse of the array or matrix to be calculated. 
            Return: inverse of matrix mat
        """
        assert np.linalg.det(mat) !=  0, "Determinant of the matrix is zero"
        return np.linalg.inv(mat)
    
    def power_smatrix(self,mat1,k,precision=10**(-10)):
        """
        It calculates the power of a real symmetric matrix.
        Attributes:
            mat1 : The matrix or array of which power is  to be calculated.
            k : value of the power
            precision: if the absolute eigenvalues below the precision 
                       value will be considered as zero 
        Return: k'th Power of symmetric matrix mat1
        """
        eigenvalues,eigenvectors,info=la.dsyev(mat1)
        flag=0
        for i in eigenvalues:
            if i < 0.0:
                flag=1
        if flag==0:
            diag=np.zeros([eigenvectors.shape[0],eigenvectors.shape[1]],\
                          dtype='float64')
        else:
            diag=np.zeros([eigenvectors.shape[0],eigenvectors.shape[1]],\
                          dtype='complex_')
        for i in range(0,eigenvectors.shape[0]):
            if abs(eigenvalues[i]) <= precision:
                diag[i,i]=0.0
                eigenvalues[i]=0.0
            if eigenvalues[i] < 0.0:
                diag[i,i]=pow(abs(eigenvalues.item(i)),k)*pow(complex(0,1),2*k)
            else:
                diag[i,i]=pow(eigenvalues.item(i),k)
        diag=np.matmul(np.matmul(eigenvectors,diag),np.transpose(eigenvectors))
        return diag
    
    def power_hmatrix(self,mat1,k,precision=10**(-10)):
        """
        It calculates the power of a Hermitian matrix.
        Attributes:
            mat1 : The matrix or array of which power is  to be calculated.
            k : value of the power
            precision: if the absolute eigenvalues below the precision 
                       value will be considered as zero 
        Return: k'th Power of Hermitian matrix mat1
        """
        eigenvalues,eigenvectors,info=la.zheev(mat1)
        flag=0
        for i in eigenvalues:
            if i < 0.0:
                flag=1
        if flag==0:
            diag=np.zeros([eigenvectors.shape[0],eigenvectors.shape[1]],\
                          dtype='float64')
        else:
            diag=np.zeros([eigenvectors.shape[0],eigenvectors.shape[1]],\
                          dtype='complex_')
        for i in range(0,eigenvectors.shape[0]):
            if abs(eigenvalues[i]) <= precision:
                diag[i,i]=0.0
                eigenvalues[i]=0.0
            if eigenvalues[i] < 0.0:
                diag[i,i]=pow(abs(eigenvalues.item(i)),k)*pow(complex(0,1),2*k)
            else:
                diag[i,i]=pow(eigenvalues.item(i),k)
        diag=np.matmul(np.matmul(eigenvectors,diag),np.conjugate(\
                       np.transpose(eigenvectors)))
        return diag  
    
    def power_gmatrix(self,mat1,k,precision=10**(-10)):
        """ 
        Calculates the power of a general non-Herimitian matrix
        Attributes:
            mat1 : The matrix or array of which power is  to be calculated.
            k : value of the power
            precision: if the absolute eigenvalues below the precision 
                       value will be considered as zero 
        Return: k'th Power of non-Hermitian matrix mat1
        """
        eigenvalues,eigenvectors=np.linalg.eig(mat1)
        diag=np.zeros([eigenvectors.shape[0],eigenvectors.shape[1]],\
                      dtype=np.complex_)
        for i in range(0,eigenvectors.shape[0]):
            if abs(eigenvalues[i]) <= precision:
                diag[i,i]=complex(0.0,0.0)
            else:
                diag[i,i]=pow(eigenvalues.item(i),k)
        
        diag=np.matmul(np.matmul(eigenvectors,diag),np.linalg.inv(eigenvectors))
        return diag        
    
    def function_smatrix(self, mat1, mode="exp",log_base=2):
        """
        It calculates the function of a real symmetric matrix.  
        Attributes:
            mat1 : The symmetric matrix of which function is to be calculated.
            mode: Primarily calculates the following,
                mode='exp': Exponential of a matrix. It is the default mode.
                mode='sin': sine of a matrix.
                mode='cos': cosine of matrix.
                mode='tan': tan of matrix.
                mode='log': Logarithm of a matrix, by default log base 2.
            log_base: base of the log function
        Return: Function of symmetric matrix mat1
        """
        
        assert np.allclose(mat1, np.matrix.transpose(mat1))==True,\
            "The matrix entered is not a symmetric matrix"
        
        assert mat1.shape[0] == mat1.shape[1],\
            "Entered matrix is not a square matrix"
        
        if mode not in ["exp","sin","cos","tan","log"]:
            raise Exception(f"Sorry, the entered mode {mode} is not available")
        
        eigenvalues,eigenvectors,info=la.dsyev(mat1)
        
        if mode == 'exp':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.exp(eigenvalues[i])
                
        if mode == 'sin':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.sin(eigenvalues[i])                

        if mode == 'cos':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.cos(eigenvalues[i])            
    
        if mode == 'tan':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.tan(eigenvalues[i])        
        
        if mode == 'log':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                assert eigenvalues[i] > 0.0,\
                "eigenvalues of the matrix are negative or zero"
                diagonal[i,i] = math.log(eigenvalues[i],log_base)
        
        return np.matmul(np.matmul(eigenvectors,diagonal),\
                         np.matrix.transpose(eigenvectors))
        
    
    def function_hmatrix(self, mat1, mode="exp",log_base=2):
        """
        It calculates the function of hermitian matrix. 
        Attributes:
            mat1 : The Hermitian matrix of which function is to be calculated.
            mode: Primarily calculates the following,
                mode='exp': Exponential of a matrix. It is the default mode.
                mode='sin': sine of a matrix.
                mode='cos': cosine of matrix.
                mode='tan': tan of matrix.
                mode='log': Logarithm of a matrix, by default log base 2.
            log_base: base of the log function
        Return: Function of Hermitian matrix mat1
        """        

        assert np.allclose(mat1, np.transpose(np.conjugate(mat1)))==True \
                              ,"The matrix entered is not a hermitian matrix"
        
        assert mat1.shape[0] == mat1.shape[1],\
            "Entered matrix is not a square matrix"
        
        if mode not in ["exp","sin","cos","tan","log"]:
            raise Exception(f"Sorry, the entered mode {mode} is not available")
        
        eigenvalues,eigenvectors,info=la.zheev(mat1)
        
        if mode == 'exp':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.exp(eigenvalues[i])
                
        if mode == 'sin':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.sin(eigenvalues[i])                

        if mode == 'cos':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.cos(eigenvalues[i])            
    
        if mode == 'tan':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = math.tan(eigenvalues[i])        
        
        if mode == 'log':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=float)
            for i in range(0,diagonal.shape[0]):
                assert eigenvalues[i] > 0.0, "eigenvalues of the matrix are negative"
                diagonal[i,i] = math.log(eigenvalues[i],log_base)
        
        return np.matmul(np.matmul(eigenvectors,diagonal),\
                         np.transpose(np.conjugate(eigenvectors)))
        
    def function_gmatrix(self, mat1, mode="exp",log_base=2):
        """ 
        It calculates the function of general diagonalizable matrix. 
        Attributes:
            mat1: The general matrix of which function is to be calculated.
            mode: Primarily calculates the following,
                Primarily calculates the following,
                mode='exp': Exponential of a matrix.
                mode='sin': sine of a matrix.
                mode='cos': cosine of matrix.
                mode='tan': tan of matrix.
                mode='log': Logarithm of a matrix, by default log base 2.
        Return: Function of general matrix mat1
        """        
        
        assert mat1.shape[0] == mat1.shape[1],\
            "Entered matrix is not a square matrix"
        
        if mode not in ["exp","sin","cos","tan","log"]:
            raise Exception(f"Sorry, the entered mode {mode} is not available")
        
        eigenvalues,eigenvectors=np.linalg.eig(mat1)
        print(eigenvalues)
        
        if mode == 'exp':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=complex)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = cmath.exp(eigenvalues[i])
                
        if mode == 'sin':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=complex)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = cmath.sin(eigenvalues[i])                

        if mode == 'cos':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=complex)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = cmath.cos(eigenvalues[i])            
    
        if mode == 'tan':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=complex)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = cmath.tan(eigenvalues[i])        
        
        if mode == 'log':
            diagonal=np.zeros((mat1.shape[0],mat1.shape[1]),dtype=complex)
            for i in range(0,diagonal.shape[0]):
                diagonal[i,i] = cmath.log(eigenvalues[i],log_base)
        assert np.linalg.det(eigenvectors) != 0, "Determinant of eigenvectors \
                                           matrix is zero"
        
        return np.matmul(np.matmul(eigenvectors,diagonal),\
                         np.linalg.inv(eigenvectors))
    
    def trace_norm_rmatrix(self,mat1, precision=10**(-13)):
        """ 
        Calculates the trace norm of a real matrix
        Attributes:
            mat1 : The matrix or array of which trace norm is to be calculated.
            precision: the absolute value of the eigenvalues below precision
                       value will be considered as zero
        Return: 
            trace_norm: trace norm of matrix mat1
        """
        eigenvalues,eigenvectors,info=la.dsyev(np.matmul(np.transpose(mat1),\
                                                         mat1))
        trace_norm=0.0
        for i in range(len(eigenvalues)):
            if abs(eigenvalues[i]) < precision:
                eigenvalues[i]=0.0
            trace_norm=trace_norm+np.sqrt(eigenvalues[i])
        
        return trace_norm

    def trace_norm_cmatrix(self,mat1, precision=10**(-13)):
        """ 
        Calculates the trace norm of a complex matrix
        Attributes:
            mat1 : The matrix or array of which trace norm is to be calculated.
            precision: the absolute value of the eigenvalues below precision
                        value will be considered as zero.
        Return: 
            trace_norm: trace norm of matrix mat1
        """
        eigenvalues,eigenvectors,info=\
        la.zheev(np.matmul(np.conjugate(np.transpose(mat1)),mat1))
        trace_norm=0.0
        for i in range(len(eigenvalues)):
            if abs(eigenvalues[i]) < precision:
                eigenvalues[i]=0.0
            trace_norm=trace_norm+np.sqrt(eigenvalues[i])
        
        return trace_norm        
        
    def hilbert_schmidt_norm_rmatrix(self,mat1, precision=10**(-13)):
        """ 
        Calculates the Hilbert-Schmidt norm of matrix of a real matrix
        Attributes:
            mat1 : The matrix or array of which Hilbert-Schmidt norm
                      is to be calculated.
            precision: tolerance value, the magnitude of eigenvalues below 
                       precision is considered zero
        Return:
            htrace_norm: Hilbert-Schmidt norm of matrix mat1.
        """
        eigenvalues,eigenvectors,info=la.dsyev(np.matmul(np.transpose(mat1),\
                                                         mat1))
        htrace_norm=0.0
        for i in range(len(eigenvalues)):
            if abs(eigenvalues[i]) < precision:
                eigenvalues[i]=0.0
            htrace_norm=htrace_norm+eigenvalues[i]
        htrace_norm=np.sqrt(htrace_norm)
        return htrace_norm

    def hilbert_schmidt_norm_cmatrix(self,mat1, precision=10**(-13)):
        """ 
        Calculates the trace norm of a complex matrix
        Attributes:
            mat1 : The matrix or array of which Hilbert-Schmidt norm 
                      is to be calculated.
            precision: tolerance value, the magnitude of eigenvalues below 
                       precision is considered zero.
        Return:
            htrace_norm: Hilbert-Schmidt norm of matrix mat1.
        """
        eigenvalues,eigenvectors,info=\
        la.zheev(np.matmul(np.conjugate(np.transpose(mat1)),mat1))
        htrace_norm=0.0
        for i in range(len(eigenvalues)):
            if abs(eigenvalues[i]) < precision:
                eigenvalues[i]=0.0
            htrace_norm=htrace_norm+eigenvalues[i]
        htrace_norm=np.sqrt(htrace_norm)
        
        return htrace_norm                
        
    def absolute_value_rmatrix(self,mat1):
        """
        Calculates the absolute value of a real matrix
        Attributes:
            mat1 : The matrix of which absolute form has to calculated.
        Return:
            res_mat: Absoulte value of matrix mat1
        """
        res_mat=self.power_smatrix(np.matmul(np.transpose(mat1),\
                                             mat1),0.50)
                
        return res_mat             
        
    def absolute_value_cmatrix(self,mat1):
        """
        Calculates the absolute value of a complex matrix
        Attributes:
            mat1 : The matrix of which absolute form has to calculated.
        Return:
            res_mat: Absoulte value of matrix mat1
        """
        res_mat=self.power_hmatrix(np.matmul(np.conjugate(np.transpose(mat1)),\
                                             mat1),0.50)           
        return res_mat         

    def hilbert_schmidt_inner_product(self,A,B):
        """ 
        Calculates the Hilbert-Schmidt inner product between matrices.
        Attributes:
            A: It is a complex or real input matrix.
            B: It is a complex or real input matrix.
        Return: Hilbert-Schmidt inner product between A and B.
        """
        return np.trace(np.matmul(np.conjugate(np.transpose(A)),B))


    def gram_schmidt_ortho_rmatrix(self,vectors):
        """
        Orthornormal set of real vectors are calculated
        Attributes:
          vectors: A matrix whose columns are non-orthogonal set real vectors
        Return:
          orthonormal_vec: A matrix whose columns are orthonormal to each other
        """
        orthonormal_vec=np.zeros((vectors.shape[0],vectors.shape[1]),
                                    dtype='float64')
        for col in range(0,vectors.shape[1]):
            if col != 0:
                for col2 in range(0,col):
                    tr=0.0
                    for row2 in range(0,vectors.shape[0]):
                        tr=tr+(orthonormal_vec[row2,col2]*vectors[row2,col])
                    orthonormal_vec[:,col]=orthonormal_vec[:,col]+\
                                               (tr*orthonormal_vec[:,col2])
                orthonormal_vec[:,col]=vectors[:,col]-orthonormal_vec[:,col]
            if col == 0:
                orthonormal_vec[:,col]=vectors[:,col]
            tr=0.0
            for row in range(0,vectors.shape[0]):
                tr=tr+(orthonormal_vec[row,col]*orthonormal_vec[row,col])
            orthonormal_vec[:,col]=orthonormal_vec[:,col]/np.sqrt(tr)
            
        return orthonormal_vec


    def gram_schmidt_ortho_cmatrix(self,vectors):
        """
        Orthornormal set of complex vectors are calculated
        Attributes:
          vectors: A matrix whose columns are non-orthogonal set 
                   complex vectors
        Return:
          orthonormal_vec: A matrix whose columns are orthonormal to each other
        """
        orthonormal_vec=np.zeros((vectors.shape[0],vectors.shape[1]),\
                                 dtype=np.complex_)
        for col in range(0,vectors.shape[1]):
            if col != 0:
                orthonormal_vec[:,col]=vectors[:,col].copy()
                for col2 in range(0,col):
                    tr=complex(0.0,0.0)
                    for row2 in range(0,vectors.shape[0]):
                        tr=tr+(np.conjugate(orthonormal_vec[row2,col2])\
                               *vectors[row2,col])
                    orthonormal_vec[:,col]=orthonormal_vec[:,col]-\
                                               (tr*\
                                                orthonormal_vec[:,col2].copy())
            if col == 0:
                orthonormal_vec[:,col]=vectors[:,col].copy()
            tr=complex(0.0,0.0)
            for row in range(0,vectors.shape[0]):
                tr=tr+(np.conjugate(orthonormal_vec[row,col])*\
                       orthonormal_vec[row,col])
            orthonormal_vec[:,col]=orthonormal_vec[:,col]/np.sqrt(tr.real)
            
        return orthonormal_vec
        
    
    
    
    
    



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "QuantumInformation/README.md").read_text()

setup(name="QuantumInformation",
     version='1.1',
     description='Quantum Information and Computation package',
     author = 'M. S. Ramkarthik and Pranay Barkataki',
     email = 'pranay.barkataki@gmail.com',
     packages=['QuantumInformation'],
     long_description=long_description,
     long_description_content_type='text/markdown',
     zip_safe=False) 
